/*
 *  i386 micro operations (included several times to generate
 *  different operand sizes)
 * 
 *  Copyright (c) 2005-2016 FAUmachine Team.
 *  Copyright (c) 2003 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * USA
 */
#define DATA_BITS (1 << (3 + SHIFT))
#define SIGN_MASK (((target_ulong)1) << (DATA_BITS - 1))

#if DATA_BITS == 8
#define SUFFIX b
#define DATA_TYPE uint8_t
#define DATA_MASK 0xff
#elif DATA_BITS == 16
#define SUFFIX w
#define DATA_TYPE uint16_t
#define DATA_MASK 0xffff
#elif DATA_BITS == 32
#define SUFFIX l
#define DATA_TYPE uint32_t
#define DATA_MASK 0xffffffff
#elif DATA_BITS == 64
#define SUFFIX q
#define DATA_TYPE uint64_t
#define DATA_MASK 0xffffffffffffffffULL
#else
#error unhandled operand size
#endif

/* dynamic flags computation */

static int glue(compute_all_div, SUFFIX)(struct cpssp *cpssp)
{
    return (0 << CPU_ZF_SHIFT);
}

static int glue(compute_c_div, SUFFIX)(struct cpssp *cpssp)
{
    return (0 << CPU_ZF_SHIFT);
}

static int glue(compute_all_add, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_src;
    src2 = cpssp->cc_dst - cpssp->cc_src;
    cf = (DATA_TYPE)cpssp->cc_dst < (DATA_TYPE)src1;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = lshift((src1 ^ src2 ^ -1) & (src1 ^ cpssp->cc_dst), 12 - DATA_BITS) & CC_O;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_add, SUFFIX)(struct cpssp *cpssp)
{
    int cf;
    target_long src1;
    src1 = cpssp->cc_src;
    cf = (DATA_TYPE)cpssp->cc_dst < (DATA_TYPE)src1;
    return cf;
}

static int glue(compute_all_adc, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_src;
    src2 = cpssp->cc_dst - cpssp->cc_src - 1;
    cf = (DATA_TYPE)cpssp->cc_dst <= (DATA_TYPE)src1;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = lshift((src1 ^ src2 ^ -1) & (src1 ^ cpssp->cc_dst), 12 - DATA_BITS) & CC_O;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_adc, SUFFIX)(struct cpssp *cpssp)
{
    int cf;
    target_long src1;
    src1 = cpssp->cc_src;
    cf = (DATA_TYPE)cpssp->cc_dst <= (DATA_TYPE)src1;
    return cf;
}

static int glue(compute_all_sub, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_dst + cpssp->cc_src;
    src2 = cpssp->cc_src;
    cf = (DATA_TYPE)src1 < (DATA_TYPE)src2;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = lshift((src1 ^ src2) & (src1 ^ cpssp->cc_dst), 12 - DATA_BITS) & CC_O;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_sub, SUFFIX)(struct cpssp *cpssp)
{
    int cf;
    target_long src1, src2;
    src1 = cpssp->cc_dst + cpssp->cc_src;
    src2 = cpssp->cc_src;
    cf = (DATA_TYPE)src1 < (DATA_TYPE)src2;
    return cf;
}

static int glue(compute_all_sbb, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_dst + cpssp->cc_src + 1;
    src2 = cpssp->cc_src;
    cf = (DATA_TYPE)src1 <= (DATA_TYPE)src2;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = lshift((src1 ^ src2) & (src1 ^ cpssp->cc_dst), 12 - DATA_BITS) & CC_O;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_sbb, SUFFIX)(struct cpssp *cpssp)
{
    int cf;
    target_long src1, src2;
    src1 = cpssp->cc_dst + cpssp->cc_src + 1;
    src2 = cpssp->cc_src;
    cf = (DATA_TYPE)src1 <= (DATA_TYPE)src2;
    return cf;
}

static int glue(compute_all_logic, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    cf = 0;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = 0;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = 0;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_logic, SUFFIX)(struct cpssp *cpssp)
{
    return 0;
}

static int glue(compute_all_inc, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_dst - 1;
    src2 = 1;
    cf = cpssp->cc_src;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = ((cpssp->cc_dst & DATA_MASK) == SIGN_MASK) << 11;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_inc, SUFFIX)(struct cpssp *cpssp)
{
    return cpssp->cc_src;
}

static int glue(compute_all_dec, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    target_long src1, src2;
    src1 = cpssp->cc_dst + 1;
    src2 = 1;
    cf = cpssp->cc_src;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = (cpssp->cc_dst ^ src1 ^ src2) & 0x10;
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = ((cpssp->cc_dst & DATA_MASK) == ((target_ulong)SIGN_MASK - 1)) << 11;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_all_shl, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    cf = (cpssp->cc_src >> (DATA_BITS - 1)) & CC_C;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = 0; /* undefined */
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    /* of is defined if shift count == 1 */
    of = lshift(cpssp->cc_src ^ cpssp->cc_dst, 12 - DATA_BITS) & CC_O;
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_shl, SUFFIX)(struct cpssp *cpssp)
{
    return (cpssp->cc_src >> (DATA_BITS - 1)) & CC_C;
}

static int glue(compute_c_sar, SUFFIX)(struct cpssp *cpssp)
{
    return cpssp->cc_src & 1;
}

static int glue(compute_all_sar, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    cf = cpssp->cc_src & 1;
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = 0; /* undefined */
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    /* of is defined if shift count == 1 */
    of = lshift(cpssp->cc_src ^ cpssp->cc_dst, 12 - DATA_BITS) & CC_O; 
    return cf | pf | af | zf | sf | of;
}

static int glue(compute_c_mul, SUFFIX)(struct cpssp *cpssp)
{
    int cf;
    cf = (cpssp->cc_src != 0);
    return cf;
}

/* NOTE: we compute the flags like the P4. On olders CPUs, only OF and
   CF are modified and it is slower to do that. */
static int glue(compute_all_mul, SUFFIX)(struct cpssp *cpssp)
{
    int cf, pf, af, zf, sf, of;
    cf = (cpssp->cc_src != 0);
    pf = CHIP_(parity_table)[(uint8_t)cpssp->cc_dst];
    af = 0; /* undefined */
    zf = ((DATA_TYPE)cpssp->cc_dst == 0) << 6;
    sf = lshift(cpssp->cc_dst, 8 - DATA_BITS) & 0x80;
    of = cf << 11;
    return cf | pf | af | zf | sf | of;
}

#undef DATA_BITS
#undef SIGN_MASK
#undef DATA_TYPE
#undef DATA_MASK
#undef SUFFIX
