/*
 * Copyright (C) 2003-2009 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include <assert.h>
#include <stdio.h>

#include "glue.h"

#include "sig_manage.h"

void
sig_manage_connect_port(
	struct sig_manage *b,
	const char *str,
	void *sig
)
{
	unsigned int nr;

	for (nr = 0; nr < b->nmembers; nr++) {
		if (! b->member[nr].f->connect) {
			continue;
		}
		(b->member[nr].f->connect)(b->member[nr].s, str, sig);
	}
}

void
sig_manage_disconnect_port(
	struct sig_manage *b,
	const char *str
)
{
	unsigned int nr;

	for (nr = 0; nr < b->nmembers; nr++) {
		if (! b->member[nr].f->disconnect) {
			continue;
		}
		(b->member[nr].f->disconnect)(b->member[nr].s, str);
	}
}

void
sig_manage_connect(
	struct sig_manage *b,
	void *s,
	const struct sig_manage_funcs *f
)
{
	assert(b);
	assert(b->type == SIG_GEN_MANAGE);
	assert(b->nmembers < sizeof(b->member) / sizeof(b->member[0]));
	assert(s);
	assert(f);

	b->member[b->nmembers].s = s;
	b->member[b->nmembers].f = f;

	b->nmembers++;
}

struct sig_manage *
sig_manage_create(const char *name)
{
	struct sig_manage *b;

	b = shm_alloc(sizeof(*b));
	assert(b);

	b->type = SIG_GEN_MANAGE;
	b->nmembers = 0;

	return b;
}

void
sig_manage_destroy(struct sig_manage *sig)
{
	assert(sig);
	assert(sig->type == SIG_GEN_MANAGE);

	shm_free(sig);
}

void
sig_manage_suspend(struct sig_manage *b, FILE *fSig)
{
	size_t size = sizeof(*b);
	
	generic_suspend(b, size, fSig);
}

void
sig_manage_resume(struct sig_manage *b, FILE *fSig)
{
	size_t size = sizeof(*b);
	
	generic_resume(b, size, fSig);
}
